"""Group data classes for the Cognite API Groups endpoint.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from .acls import (
    Acl,
    AgentsAcl,
    AnalyticsAcl,
    AnnotationsAcl,
    AppConfigAcl,
    AssetsAcl,
    AuditlogAcl,
    DataModelInstancesAcl,
    DataModelsAcl,
    DataSetsAcl,
    DiagramParsingAcl,
    DigitalTwinAcl,
    DocumentFeedbackAcl,
    DocumentPipelinesAcl,
    EntityMatchingAcl,
    EventsAcl,
    ExperimentsAcl,
    ExtractionConfigsAcl,
    ExtractionPipelinesAcl,
    ExtractionRunsAcl,
    FilePipelinesAcl,
    FilesAcl,
    FunctionsAcl,
    GeospatialAcl,
    GeospatialCrsAcl,
    GroupsAcl,
    HostedExtractorsAcl,
    LabelsAcl,
    LegacyGenericsAcl,
    LegacyModelHostingAcl,
    LocationFiltersAcl,
    MonitoringTasksAcl,
    NotificationsAcl,
    PipelinesAcl,
    PostgresGatewayAcl,
    ProjectsAcl,
    RawAcl,
    RelationshipsAcl,
    RoboticsAcl,
    SAPWritebackAcl,
    SAPWritebackRequestsAcl,
    ScheduledCalculationsAcl,
    SecurityCategoriesAcl,
    SeismicAcl,
    SequencesAcl,
    SessionsAcl,
    SimulatorsAcl,
    StreamRecordsAcl,
    StreamsAcl,
    TemplateGroupsAcl,
    TemplateInstancesAcl,
    ThreeDAcl,
    TimeSeriesAcl,
    TimeSeriesSubscriptionsAcl,
    TransformationsAcl,
    TypesAcl,
    UnknownAcl,
    UserProfilesAcl,
    VideoStreamingAcl,
    VisionModelAcl,
    WellsAcl,
    WorkflowOrchestrationAcl,
)
from .capability import (
    GroupCapability,
    ProjectUrlNames,
)
from .group import (
    Group,
    GroupAttributes,
    GroupRequest,
    GroupResponse,
    TokenAttributes,
)
from .scopes import (
    AllScope,
    AppConfigScope,
    AssetRootIDScope,
    CurrentUserScope,
    DataSetScope,
    ExperimentScope,
    ExtractionPipelineScope,
    IDScope,
    IDScopeLowerCase,
    InstancesScope,
    PartitionScope,
    PostgresGatewayUsersScope,
    Scope,
    ScopeDefinition,
    SpaceIDScope,
    TableScope,
)

__all__ = [
    "Acl",
    "AgentsAcl",
    "AllScope",
    "AnalyticsAcl",
    "AnnotationsAcl",
    "AppConfigAcl",
    "AppConfigScope",
    "AssetRootIDScope",
    "AssetsAcl",
    "AuditlogAcl",
    "CurrentUserScope",
    "DataModelInstancesAcl",
    "DataModelsAcl",
    "DataSetScope",
    "DataSetsAcl",
    "DiagramParsingAcl",
    "DigitalTwinAcl",
    "DocumentFeedbackAcl",
    "DocumentPipelinesAcl",
    "EntityMatchingAcl",
    "EventsAcl",
    "ExperimentScope",
    "ExperimentsAcl",
    "ExtractionConfigsAcl",
    "ExtractionPipelineScope",
    "ExtractionPipelinesAcl",
    "ExtractionRunsAcl",
    "FilePipelinesAcl",
    "FilesAcl",
    "FunctionsAcl",
    "GeospatialAcl",
    "GeospatialCrsAcl",
    "Group",
    "GroupAttributes",
    "GroupCapability",
    "GroupRequest",
    "GroupResponse",
    "GroupsAcl",
    "HostedExtractorsAcl",
    "IDScope",
    "IDScopeLowerCase",
    "InstancesScope",
    "LabelsAcl",
    "LegacyGenericsAcl",
    "LegacyModelHostingAcl",
    "LocationFiltersAcl",
    "MonitoringTasksAcl",
    "NotificationsAcl",
    "PartitionScope",
    "PipelinesAcl",
    "PostgresGatewayAcl",
    "PostgresGatewayUsersScope",
    "ProjectUrlNames",
    "ProjectsAcl",
    "RawAcl",
    "RelationshipsAcl",
    "RoboticsAcl",
    "SAPWritebackAcl",
    "SAPWritebackRequestsAcl",
    "ScheduledCalculationsAcl",
    "Scope",
    "ScopeDefinition",
    "SecurityCategoriesAcl",
    "SeismicAcl",
    "SequencesAcl",
    "SessionsAcl",
    "SimulatorsAcl",
    "SpaceIDScope",
    "StreamRecordsAcl",
    "StreamsAcl",
    "TableScope",
    "TemplateGroupsAcl",
    "TemplateInstancesAcl",
    "ThreeDAcl",
    "TimeSeriesAcl",
    "TimeSeriesSubscriptionsAcl",
    "TokenAttributes",
    "TransformationsAcl",
    "TypesAcl",
    "UnknownAcl",
    "UserProfilesAcl",
    "VideoStreamingAcl",
    "VisionModelAcl",
    "WellsAcl",
    "WorkflowOrchestrationAcl",
]
